@@ -132,6 +132,7 @@ def batchify(items, wrapper=np.array):
 
 
 class Dataset(metaclass=ABCMeta):
+    _parent = None
     """The base class for any dataset, provides the and batches methods from
     yield_keys() and query_item(key)
 
@@ -162,10 +163,21 @@ class Dataset(metaclass=ABCMeta):
     def query(self, keys, wrapper=np.array):
         return batchify([self.query_item(key) for key in keys], wrapper=wrapper)
 
-    def batches(self, keys, batch_size, wrapper=np.array, drop_incomplete=False):
+    def batches(self, batch_size: int, keys=None, wrapper=np.array, drop_incomplete=False):
+        assert isinstance(batch_size, int), "`batches` signature has changed: keys is now given after batch size and " \
+            "defaults to the whole datasets if `None` is given."
+        keys = keys or self.keys
         for key_chunk in chunkify(keys, n=batch_size, drop_incomplete=drop_incomplete):
             yield key_chunk, self.query(key_chunk, wrapper)
 
+    def __len__(self):
+        keys = self.keys
+        if self._parent is not None:
+            return len(self._parent)
+        if isinstance(keys, _CompleteDatasetKeys):
+            return len(keys)
+        raise RuntimeError("Impossible to compute dataset lenght without computing it.")
+
     @property
     def parent(self):
         return self._parent
@@ -309,6 +321,9 @@ class DictDataset(Dataset):
     def query_item(self, key):
         return self.dic[key]
 
+    def __len__(self):
+        return len(self.dic)
+
 
 class FilteredDataset(AugmentedDataset):
     def __init__(self, parent, predicate, keep_positive=True):
@@ -316,10 +331,10 @@ class FilteredDataset(AugmentedDataset):
         self.predicate = predicate
         self.keep_positive = keep_positive
 
-    def augment(self, key, item):
-        truth_value = self.predicate(key, item)
+    def augment(self, root_key, root_item):
+        truth_value = self.predicate(root_key, root_item)
         if truth_value is self.keep_positive:
-            yield (key, item)
+            yield (root_key, root_item)
         else:
             assert truth_value is (not self.keep_positive), (
                 "Predicate {!r} should return a boolean value"
@@ -346,6 +361,7 @@ class PickledDataset(Dataset):
             X, Y = pd.query(pd.keys())
             model.fit(X, Y)
     """
+    _last_fork = 0
     @staticmethod
     def create(dataset, file_handler, yield_keys_wrapper=None, keys=None):
         if isinstance(file_handler, str):
@@ -378,45 +394,59 @@ class PickledDataset(Dataset):
         index_location ^= 1 << 65
         pickler.dump(index_location)
 
-    def __init__(self, file_handler):
-        if isinstance(file_handler, str):
-            file_handler = open(file_handler, "rb")
-        self.file_handler = file_handler
-        self.unpickler = unpickler = Unpickler(file_handler)
-
+    def __init__(self, filename):
         # load the index offset then the index
-        file_handler.seek(0)
-        index_location = unpickler.load()
-        index_location ^= 1 << 65
-        file_handler.seek(index_location)
-        self.index = unpickler.load()
-        unpickler.memo.clear()
-        # try to load the context if any
-        try:
-            self._context = ChainMap(unpickler.load())
-            unpickler.memo.clear()
-        except EOFError:
-            pass
-
         self.lock = threading.Lock()
+        with self.lock:
+            self._open(filename)
+
+            self.file_handler.seek(0)
+            index_location = self.unpickler.load()
+            index_location ^= 1 << 65
+            self.file_handler.seek(index_location)
+            self.index = self.unpickler.load()
+            self.unpickler.memo.clear()
+            # try to load the context if any
+            try:
+                self._context = ChainMap(self.unpickler.load())
+                self.unpickler.memo.clear()
+            except EOFError:
+                pass
+
+    def _open(self, filename):
+        self.file_handler = open(filename, "rb")
+        self.unpickler = Unpickler(self.file_handler)
+        self._last_fork = PickledDataset._last_fork
 
     def __getstate__(self):
         return (self.file_handler.name,)
 
     def __setstate__(self, state):
-        self.__init__(*state)
+        file_handler, = state
+        self.__init__(file_handler)
 
     def yield_keys(self):
         return self.index.keys()
 
     def query_item(self, key):
-        self.lock.acquire()
-        self.file_handler.seek(self.index[key])
-        ret = self.unpickler.load()
-        self.unpickler.memo.clear()
-        self.lock.release()
+        with self.lock:
+            while self._last_fork < PickledDataset._last_fork:
+                self._open(self.file_handler.name)
+
+            self.file_handler.seek(self.index[key])
+            ret = self.unpickler.load()
+            self.unpickler.memo.clear()
         return ret
 
+    def __len__(self):
+        return len(self.index)
+
+    @staticmethod
+    def _inc_fork():
+        PickledDataset._last_fork += 1
+
+os.register_at_fork(after_in_child=PickledDataset._inc_fork)
+
 
 class DiffReason(Exception):
     pass
